/*
 * nbtscanner -- a tool for scanning large networks for SMB servers.
 *
 * configfile.c: Functions for parsing the configuration file.
 * Copyright (C) 2000 Steinar H. Gunderson
 *
 * Large amounts of code adapted from Samba (http://www.samba.org/)
 * Copyright (C) Andrew Tridgell 1994-1998, and others.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "configfile.h"
#include "nbtscanner.h"
#include "mysql_interface.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

int num_retries;
int retry_time;
int delay_time;
int verbosity;

int scan_interval;
int scan_wait;

struct in_addr scanrange[NUM_RANGES];
int scanrangesize[NUM_RANGES];
int ranges;

int use_mysql;
char mysql_host[64];
char mysql_username[32];
char mysql_password[32];

void parse_configfile()
{
	FILE *configfile = fopen("nbtscanner.conf", "r");
	int lineno = 0;

	if (configfile == NULL) {
		perror("nbtscanner.conf");
		exit(1);
	}

	/* defaults */
	num_retries = 3;
	retry_time = 5000;
	delay_time = 100;
	verbosity = 2;

	ranges = 0;

	use_mysql = 0;
	strcpy(mysql_host, "no_host_given");
	strcpy(mysql_username, "no_username_given");
	strcpy(mysql_password, "no_password_given");

	while (!feof(configfile)) {
		char buf[1024];
	
		lineno++;

		fgets(buf, 1024, configfile);
		while (buf[strlen(buf) - 1] == '\n' || buf[strlen(buf) - 1] == '\r') {
			buf[strlen(buf) - 1] = '\0';
		}

		if (buf[0] == '#' || buf[0] == '\0') {
			/* comment or blank line */
			continue;
		}
		if (strspn(buf, " \t") == strlen(buf)) {
			/* blank line */
			continue;
		}
		
		parse_line(buf, lineno);
	}

	fclose(configfile);
}

void parse_line(char *buf, int lineno)
{
	char *ptr = strchr(buf, '=');

	if (ptr == NULL || strchr(ptr + 1, '=') != NULL) {
		line_error(lineno, "Malformed line (must be exactly one `=' sign)");
	}

	/* split the line into a keyword and a value */
	ptr[0] = 0;
	ptr++;

	/*
	 * trim any trailing spaces, make sure there are no spaces in
	 * the value
	 */
	{
		char *ptr2 = strchr(ptr, ' ');
		if (ptr2 != NULL) {
			/* must only be spaces from here */
			if (strspn(ptr2 + 1, " ") != strlen(ptr2 + 1)) {
				line_error(lineno, "Malformed value (can't have any spaces)");
			}

			ptr2[0] = 0;
		}
	}

	/* make sure both the parameter and the value is non-empty */
	if (buf[0] == '\0') {
		line_error(lineno, "Malformed keyword (can't be empty)");
	}
	if (ptr[0] == '\0') {
		line_error(lineno, "Malformed value (can't be empty)");
	}

	/* finally pass it on to the keyword checker */
	parse_keyword(buf, ptr, lineno);
}

void parse_keyword(char *keyword, char *value, int lineno)
{
	if (strcasecmp(keyword, "num_retries") == 0) {
		parse_int(value, &num_retries, lineno);
	} else if (strcasecmp(keyword, "retry_time") == 0) {
		parse_int(value, &retry_time, lineno);
	} else if (strcasecmp(keyword, "delay_time") == 0) {
		parse_int(value, &delay_time, lineno);
	} else if (strcasecmp(keyword, "verbosity") == 0) {
		parse_int(value, &verbosity, lineno);
	} else if (strcasecmp(keyword, "scan_interval") == 0) {
		parse_int(value, &scan_interval, lineno);
	} else if (strcasecmp(keyword, "scan_wait") == 0) {
		parse_int(value, &scan_wait, lineno);
	} else if (strcasecmp(keyword, "use_mysql") == 0) {
		parse_int(value, &use_mysql, lineno);
	} else if (strcasecmp(keyword, "mysql_host") == 0) {
		strcpy(mysql_host, value);
	} else if (strcasecmp(keyword, "mysql_username") == 0) {
		strcpy(mysql_username, value);
	} else if (strcasecmp(keyword, "mysql_password") == 0) {
		strcpy(mysql_password, value);
	} else if (strcasecmp(keyword, "range") == 0) {
		parse_range(value, lineno);
	} else {
		char buf[1024];
		sprintf(buf, "Unknown keyword `%s'", keyword);
		line_error(lineno, buf);
	}
}

void parse_int(char *string, int *retval, int lineno)
{
	if (strspn(string, "0123456789") != strlen(string)) {
		char buf[1024];
		sprintf(buf, "`%s' isn't a positive integer", string);
		line_error(lineno, buf);
	}

	*retval = atoi(string);
}

void parse_range(char *string, int lineno)
{
	/* exactly one / is allowed */
	char *ptr = strchr(string, '/');
	struct in_addr in;
	int rangesize;

	if (ptr != NULL && strchr(ptr + 1, '/') != NULL) {
		line_error(lineno, "Malformed range (only one `/' allowed)"); 
 	}

	/* lots and lots of code :-) */
	if (ptr == NULL) {		/* single IP */
		rangesize = 32;
	} else {			/* range */
		ptr[0] = 0;
		ptr++;

		/* find out whether this is a pure integer or not */
		if (strspn(ptr, "0123456789") == strlen(ptr)) {
			rangesize = atoi(ptr);
			if (rangesize < 0 || rangesize > 32) {
				line_error(lineno, "Malformed range (significant bits must be between 0 and 32, inclusive)");
			}
		} else if (strspn(ptr, "0123456789.") == strlen(ptr)) {
			/* IP address... */
			struct in_addr netmask;
			int found, i;

			if (inet_aton(ptr, &netmask) == 0) {
				line_error(lineno, "Malformed range (malformed netmask)");
			}

			/* find the rightmost 1-bit (by finding the leftmost 0-bit) */
			found = 0;
			for (i = 31; i >= 0; i--) {
				if ((netmask.s_addr & (1 << i)) == 0) {
					found = 1;
					break;
				}
			}
			if (found == 0) {
				rangesize = 32;		/* 255.255.255.255 */
			} else {
				unsigned int clean_mask = htonl(~((1 << (31-i)) - 1));

				rangesize = i + 1;

				/* verify that this is a `clean' netmask */
				if (netmask.s_addr != clean_mask) {
					line_error(lineno, "Malformed range (netmask must be `clean')");
				}
			}
		} else {
			line_error(lineno, "Malformed range (malformed netmask)");
			rangesize = 32;	/* will never reach this, but for gcc */
		}
	}

	/* try to scan the IP part */
	if (inet_aton(string, &in) == 0) {
		line_error(lineno, "Malformed range (malformed IP address)");
	}

	/* check that the IP address and the netmask match */
	if (rangesize != 32) {
		/* not really the netmask, actually it's the inverted netmask */
		unsigned int netmask = htonl((1 << (31-rangesize)) - 1);	

		if ((in.s_addr & netmask) != 0) {
			line_error(lineno, "Malformed range (IP address doesn't match netmask)");
		}
	}

	memcpy((char *)(&(scanrange[ranges])), (char *)(&in), sizeof(in));
	scanrangesize[ranges++] = rangesize;
}

void line_error(int lineno, char *err_str)
{
	fprintf(stderr, "nbtscanner.conf:%d: %s\n", lineno, err_str);
	exit(1);
}
