/*
 * nbtscanner -- a tool for scanning large networks for SMB servers.
 *
 * id_list.c: a singly linked list for keeping track of the packet IDs.
 * Copyright (C) 2000 Steinar H. Gunderson
 *
 * Large amounts of code adapted from Samba (http://www.samba.org/)
 * Copyright (C) Andrew Tridgell 1994-1998, and others.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "id_list.h"
#include "configfile.h"
#include "util.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>

struct id_entry ids[65536];
struct id_list_entry *head_free_id;
int num_free_ids;

void id_list_init()
{
	int i, start_num;
	
	num_free_ids = 0;
	head_free_id = NULL;

	/* we want to start at some halfway random number instead of 0 */
	start_num = (time(NULL) + getpid()) & 0xffff;

	for (i = 65535; i >= 0; i--) {
		int ii = (i + start_num) & 0xffff;

		ids[ii].id = ii;
		ids[ii].in_use = 0;

		id_free(&ids[ii]);
	}
}

void id_list_destroy()
{
	while (head_free_id) {
		struct id_list_entry *next = head_free_id->next;
		free(head_free_id);
		head_free_id = next;
	}
	num_free_ids = 0;
}

void id_free(struct id_entry *i)
{
	struct id_list_entry *ie = (struct id_list_entry *)malloc(sizeof(struct id_list_entry));

	if (ie == NULL) {
		perror("malloc()");
		exit(1);
	}

	i->in_use = 0;
	ie->id = i;

	/* insert at the beginning of the linked list */
	ie->next = head_free_id;
	head_free_id = ie;

	num_free_ids++;
}

/* mark as free, but don't delete until desperation cleanup */
void id_mark_free(struct id_entry *i)
{
	i->in_use = 2;
	num_free_ids++;
}

struct id_entry *id_get_free_id()
{
	struct id_list_entry *ie = head_free_id;
	struct id_entry *i;

	if (ie == NULL) {
		return NULL;
	} else {
		i = ie->id;
		head_free_id = ie->next;
		free(ie);

		i->in_use = 1;
		num_free_ids--;
		
		return i;
	}
}

/* desperation time :-) */
int id_cleanup(int timeout_ms)
{
	int i;
	int num_freed = 0;
	struct timeval now;
	gettimeofday(&now, NULL);

	for (i = 0; i < 65536; i++) {
		if (ids[i].in_use == 2) {
			id_free(&ids[i]);
			num_freed++;
			num_free_ids--;		/* was already counted */

			continue;
		}
		if (mydifftime(ids[i].timestamp, now) > timeout_ms &&
		    ids[i].retries == num_retries) {
			id_free(&ids[i]);
			num_freed++;

			continue;
		}
	}

	return num_freed;
}

int get_num_free_ids()
{
	return num_free_ids;
}
