/*
 * nbtscanner -- a tool for scanning large networks for SMB servers.
 *
 * nbtscanner.c: nbtscanner main body. 
 * Copyright (C) 2000 Steinar H. Gunderson
 *
 * Large amounts of code adapted from Samba (http://www.samba.org/)
 * Copyright (C) Andrew Tridgell 1994-1998, and others.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <stdlib.h>

#include "nbtscanner.h"
#include "mysql_interface.h"
#include "configfile.h"
#include "nmb.h"
#include "socket.h"
#include "stats.h"
#include "packet.h"
#include "byteorder.h"
#include "util.h"
#include "id_list.h"

/* one global variable -- the UDP socket */
int server_fd;

void interpret_node_status(char *p, struct in_addr in)
{
	int numnames = CVAL(p,0);
	char hostname[32], fileservername[32], group[32];

	/*
	 * hopefully this will be enough to identify the host as unknown
	 * (note that this is longer than the 16 char maximum NetBIOS
	 * name length)
	 */
	strcpy(hostname, "-unknown-nbtscanner-");
	strcpy(fileservername, "-unknown-nbtscanner-");
	strcpy(group, "-unknown-nbtscanner-");

	p += 1;
	while (numnames--) {
		char qname[17];
		int type;
		int i;

		strncpy(qname,p,15);
		qname[15] = 0;
		type = CVAL(p,15);
		p += 16;

		for (i = strlen(qname); --i >= 0; ) {
			if (!isprint((int)qname[i])) qname[i] = '.';
		}

		/*
		 * this is code duplication (see parse_nmb_name in nmb.c) 
		 * but it obviously isn't done in our case :-)
		 */
		for (i = strlen(qname); --i >= 0; ) {
			if (qname[i] == ' ') {
				qname[i] = 0;
			} else {
				break;
			}
		}

		/* <ACTIVE>, non-<GROUP> server hostname (type 0x00) */ 
		if ((p[0] & 0x80) == 0x00 && ((p[0] & 0x1f) == 0x04) && type == 0x00) {
			strcpy(hostname, qname);
		}
		/* <ACTIVE>, non-<GROUP> server hostname (type 0x20) */ 
		if ((p[0] & 0x80) == 0x00 && ((p[0] & 0x1f) == 0x04) && type == 0x20) {
			strcpy(fileservername, qname);
		}
		/* <ACTIVE>, <GROUP> group (type 0x00) */
		if ((p[0] & 0x80) == 0x80 && ((p[0] & 0x1f) == 0x04) && type == 0x00) {
			strcpy(group, qname);
		}

		p += 2;
	}

	if (use_mysql) {
		add_record_mysql(inet_ntoa(in), hostname, fileservername, group);
	} else {
		printf("%s,%s,%s,%s\n", inet_ntoa(in), hostname, fileservername, group);
	}
}

void resend_timed_out_packets(int fd)
{
	int i;
	struct timeval now;
	struct nmb_packet nmb;

	gettimeofday(&now, NULL);

	for (i = 0; i < 65536; i++) {
		if (ids[i].in_use == 1 &&
		    mydifftime(ids[i].timestamp, now) > retry_time) {
			if (ids[i].retries >= num_retries) {
				id_mark_free(&ids[i]);

				num_failed++;
				if (use_mysql) {
					delete_record_mysql(inet_ntoa(ids[i].sin));
				}
				print_stats(0);
			} else {
				gettimeofday(&(ids[i].timestamp), NULL);
				ids[i].retries++;
			        num_sent_total++;
	
				build_nbt_packet(&nmb, i);
				send_packet(&nmb, ids[i].sin, 137, fd);
				recv_nbt_packets(fd);	/* delay and empty buffer */
			        print_stats(0);
			}
		}
	}
}

void send_nbt_packet(int fd, struct in_addr to_ip)
{
	struct nmb_packet nmb;
	static int name_trn_id = 0;
	struct id_entry *i = id_get_free_id();

	if (i == NULL) {
		/* uh-oh... try to remove timed-out IDs */
		if (id_cleanup(retry_time + 1000) == 0 ||
			(i=id_get_free_id()) == NULL) {
			if (verbosity >= 1) {
				fprintf(stderr, "No free NBT IDs! Reduce "
						"the number of hosts, the "
						"number\n");
				fprintf(stderr, "of retries, or the scanning "
						"speed. nbtscanner is now "
						"exiting.\n");
			}
			exit(1);
		}
	}
	name_trn_id = i->id;
	i->sin = to_ip;
	gettimeofday(&(i->timestamp), NULL);
	i->retries = 0;

	build_nbt_packet(&nmb, name_trn_id);

	/* FIXME: what to do with sending errors? */
	send_packet(&nmb, to_ip, 137, fd);

	num_sent++;
	num_sent_total++;
	print_stats(0);
}

/*
 * This function not only receives and handles any NBT replies, but
 * also guarantees a delay of minimum delay_time.
 */
void recv_nbt_packets(int fd)
{
	struct nmb_packet *nmb;
	struct in_addr in;
	struct timeval start;
	int dt = delay_time;

	gettimeofday(&start, NULL);

	while (dt > 0) {
		struct timeval now;

		print_stats(0);

		gettimeofday(&now, NULL);
		dt = delay_time - mydifftime(start, now);

		if (dt < 0) break;

		nmb = receive_packet(fd, dt, &in);
		if (nmb) {
			struct id_entry *i = &ids[nmb->header.name_trn_id];
	
			if (!nmb->header.response) {
				/* It's not for us */
				free_nmb_packet(nmb);
				continue;
			}
			if (i->in_use == 0) {
				if (verbosity < 2) continue;

				fprintf(stderr, "Warning: received NBT "
						"response for unused "
						"id %04x!\n",
					i->id);
				continue;
			}
			if (i->in_use == 2) {
				if (verbosity < 2) continue;

				fprintf(stderr, "Warning: received "
						"duplicate NBT response "
						"for id %04x!\n",	
					i->id);
				continue;
			}
			if (in.s_addr != i->sin.s_addr) {
				char got_addr[32], want_addr[32];
	
				if (verbosity < 2) continue;

				strcpy(got_addr, inet_ntoa(in));
				strcpy(want_addr, inet_ntoa(i->sin));
	
				fprintf(stderr, "Warning: received NBT "
						"response for id %04x "
						"from %s instead of %s!\n",
					i->id, got_addr, want_addr);
				continue;
			}

			if (nmb->header.opcode != 0 ||
			    nmb->header.nm_flags.bcast ||
			    nmb->header.rcode ||
			    !nmb->header.ancount ||
			    nmb->answers->rr_type != 0x21) {
				/* XXXX what do we do with this? could be a redirect, but
				   we'll discard it for the moment */
				free_nmb_packet(nmb);
				continue;
			}

			num_recv++;
			num_recv_retries = i->retries;
	
			interpret_node_status(&nmb->answers->rdata[0], in);
			free_nmb_packet(nmb);
	
			id_mark_free(i);
		}
	}
}	

void scan_range(struct in_addr ip, int rangesize)
{
	if (rangesize == 32) {
		/* scan a single IP */
		send_nbt_packet(server_fd, ip);
		recv_nbt_packets(server_fd);
	} else if (rangesize == 31) {
		/* scan two IPs */
		send_nbt_packet(server_fd, ip);
		recv_nbt_packets(server_fd);

		ip.s_addr = htonl(ntohl(ip.s_addr) + 1);
		send_nbt_packet(server_fd, ip);
		recv_nbt_packets(server_fd);
	} else {
		/* scan a range of IPs, minus the top and bottom one */
		int i;

		ip.s_addr = htonl(ntohl(ip.s_addr) + 1);	/* skip bottom */

		for (i = 0; i < (1 << (32-rangesize)) - 2; i++) {
			send_nbt_packet(server_fd, ip);
			recv_nbt_packets(server_fd);

			ip.s_addr = htonl(ntohl(ip.s_addr) + 1);
		}
	}
}

int main(int argc, char *argv[])
{
	int fd = open_sockets();
	int i;

	server_fd = fd;
	parse_configfile();

	if (use_mysql) {
		init_mysql(mysql_host, mysql_username, mysql_password);
	}

	for ( ;; ) {
		struct timeval start, now;
		unsigned int delay_ms;

		id_list_init();
		init_stats();

		gettimeofday(&start, NULL);

		for (i = 0; i < ranges; i++) {
			scan_range(scanrange[i], scanrangesize[i]);
		}
	
		/*
		 * receive answers and resend packets until all answers have
		 * been received, or all packets have been timed out
		 */
		status = RETRYING;
		while (get_num_free_ids() != 65536) {
			recv_nbt_packets(fd);
			resend_timed_out_packets(fd);
		}

		if (verbosity >= 3) {
			print_stats(1);
			fprintf(stderr, "\n");
		}

		id_list_destroy();

		gettimeofday(&now, NULL);

		delay_ms = scan_interval - mydifftime(start, now);
		if (scan_wait > delay_ms) delay_ms = scan_wait;

		return 0;

		if (verbosity >= 3) {
			fprintf(stderr, "Scan took %d ms, waiting %d ms for next scan...\r",
				mydifftime(start, now), delay_ms);
			usleep(delay_ms * 1000);
		}
	}
	/* finish_mysql(); */
	return 0;
}
