/*
 * nbtscanner -- a tool for scanning large networks for SMB servers.
 *
 * nmb.c: Most functions directly related to the NMB protocol.
 *
 * Large amounts of code adapted from Samba (http://www.samba.org/)
 * Copyright (C) Andrew Tridgell 1994-1998, and others.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>

#include "nmb.h"
#include "packet.h"
#include "byteorder.h"
#include "util.h"

void build_nbt_packet(struct nmb_packet *nmb, int name_trn_id)
{
	/* this could in theory be optimized, but I don't think it's worth it */
	nmb->header.name_trn_id = name_trn_id;
	nmb->header.opcode = 0;
	nmb->header.response = 0;
	nmb->header.nm_flags.bcast = 0;
	nmb->header.nm_flags.recursion_available = 0;
	nmb->header.nm_flags.recursion_desired = 0;
	nmb->header.nm_flags.trunc = 0;
	nmb->header.nm_flags.authoritative = 0;
	nmb->header.rcode = 0;
	nmb->header.qdcount = 1;
	nmb->header.ancount = 0;
	nmb->header.nscount = 0;
	nmb->header.arcount = 0;

	make_nmb_name(&(nmb->question.question_name), "*", 0x0, "");

	nmb->question.question_type = 0x21;
	nmb->question.question_class = 0x1;
}

int build_nmb(char *buf, struct nmb_packet *nmb)
{
	unsigned char *ubuf = (unsigned char *)buf;
	int offset=0;

	/* put in the header */
	RSSVAL(ubuf,offset, nmb->header.name_trn_id);
	ubuf[offset+2] = (nmb->header.opcode & 0xF) << 3;
	if (nmb->header.response) ubuf[offset+2] |= (1<<7);
	if (nmb->header.nm_flags.authoritative &&
	    nmb->header.response) ubuf[offset+2] |= 0x4;
	if (nmb->header.nm_flags.trunc) ubuf[offset+2] |= 0x2;
	if (nmb->header.nm_flags.recursion_desired) ubuf[offset+2] |= 0x1;
	if (nmb->header.nm_flags.recursion_available &&
	    nmb->header.response) ubuf[offset+3] |= 0x80;
	if (nmb->header.nm_flags.bcast) ubuf[offset+3] |= 0x10;
	ubuf[offset+3] |= (nmb->header.rcode & 0xF);

	RSSVAL(ubuf, offset+4, nmb->header.qdcount);
	RSSVAL(ubuf, offset+6, nmb->header.ancount);
	RSSVAL(ubuf, offset+8, nmb->header.nscount);
	RSSVAL(ubuf, offset+10, nmb->header.arcount);

	offset += 12;
	if (nmb->header.qdcount) {
		/* XXXX this doesn't handle a qdcount of > 1 */
		offset += put_nmb_name((char *)ubuf, offset, &nmb->question.question_name);
		RSSVAL(ubuf,offset, nmb->question.question_type);
		RSSVAL(ubuf,offset+2, nmb->question.question_class);
		offset += 4;
	}
	return offset;
}

int parse_nmb(char *inbuf, int length, struct nmb_packet *nmb)
{
	int nm_flags, offset;

	memset((char *)nmb, 0, sizeof(*nmb));

	if (length < 12) return(0);

	/* parse the header */
	nmb->header.name_trn_id = RSVAL(inbuf, 0);

	nmb->header.opcode = (CVAL(inbuf, 2) >> 3) & 0xF;
	nmb->header.response = ((CVAL(inbuf, 2) >> 7) & 1) ? 1 : 0;
	nm_flags = ((CVAL(inbuf, 2) & 0x7) << 4) + (CVAL(inbuf, 3) >> 4);
	nmb->header.nm_flags.bcast = (nm_flags & 1) ? 1 : 0;
	nmb->header.nm_flags.recursion_available = (nm_flags & 8) ? 1 : 0;
	nmb->header.nm_flags.recursion_desired = (nm_flags & 0x10) ? 1 : 0;
	nmb->header.nm_flags.trunc = (nm_flags & 0x20) ? 1 : 0;
	nmb->header.nm_flags.authoritative = (nm_flags & 0x40) ? 1 : 0;
	nmb->header.rcode = CVAL(inbuf, 3) & 0xF;
	nmb->header.qdcount = RSVAL(inbuf, 4);
	nmb->header.ancount = RSVAL(inbuf, 6);
	nmb->header.nscount = RSVAL(inbuf, 8);
	nmb->header.arcount = RSVAL(inbuf, 10);

	if (nmb->header.qdcount) {
		offset = parse_nmb_name(inbuf, 12, length, &nmb->question.question_name);
		if (!offset) return 0;

		if (length - (12+offset) < 4) return 0;
		nmb->question.question_type = RSVAL(inbuf, 12+offset);
		nmb->question.question_class = RSVAL(inbuf, 12+offset+2);

		offset += 12+4;
	} else {
		offset = 12;
	}

	/* and any resource records */
	if (nmb->header.ancount &&
		!parse_alloc_res_rec(inbuf, &offset, length, &nmb->answers,
				     nmb->header.ancount)) {
		return 0;
	}

	if (nmb->header.nscount &&
		!parse_alloc_res_rec(inbuf, &offset, length, &nmb->nsrecs,
				     nmb->header.nscount)) {
		return 0;
	}

	if (nmb->header.arcount &&
		!parse_alloc_res_rec(inbuf, &offset, length, &nmb->additional,
				     nmb->header.arcount)) {
		return 0;
	}

	return 1;
}

void make_nmb_name(struct nmb_name *n, const char *name, int type, const char *this_scope)
{
	memset((char *)n, 0, sizeof(struct nmb_name));
	strncpy(n->name, name, 15);
	strupper(n->name);
	n->name_type = (unsigned int)type & 0xFF;
	strncpy(n->scope, this_scope, 63);
	strupper(n->scope);
}

int parse_nmb_name(char *inbuf, int offset, int length, struct nmb_name *name)
{
	int m, n = 0;
	unsigned char *ubuf = (unsigned char *)inbuf;
	int ret = 0;
	int got_pointer = 0;
	int loop_count = 0;

	if (length - offset < 2)
		return 0;

	/* handle initial name pointers */
	if (!handle_name_ptrs(ubuf, &offset, length, &got_pointer, &ret))
		return 0;

	m = ubuf[offset];

	if (!m)
		return 0;
	if ((m & 0xC0) || offset+m+2 > length)
		return 0;

	memset((char *)name, 0, sizeof(*name));

	/* the "compressed" part */
	if (!got_pointer)
		ret += m + 2;
	offset++;

	while (m > 0) {
		unsigned char c1,c2;
		c1 = ubuf[offset++]-'A';
		c2 = ubuf[offset++]-'A';
		if ((c1 & 0xF0) || (c2 & 0xF0) || (n > sizeof(name->name)-1))
			return 0;
		name->name[n++] = (c1<<4) | c2;
		m -= 2;
	}
	name->name[n] = 0;

	if (n==16) {
		/* parse out the name type,
		   its always in the 16th byte of the name */
		name->name_type = ((unsigned char)name->name[15]) & 0xff;

		/* remove trailing spaces */
		name->name[15] = 0;
		n = 14;
		while (n && name->name[n]==' ')
			name->name[n--] = 0;
	}

	/* now the domain parts (if any) */
	n = 0;
	while (ubuf[offset]) {
		/* we can have pointers within the domain part as well */
		if (!handle_name_ptrs(ubuf,&offset,length,&got_pointer,&ret))
			return 0;

		m = ubuf[offset];
		/*
		 * Don't allow null domain parts.
		 */
		if (!m)
			return 0;
		if (!got_pointer)
			ret += m+1;
		if (n)
			name->scope[n++] = '.';
		if (m + 2 + offset > length || n + m + 1 > sizeof(name->scope))
			return 0;
		offset++;
		while (m--)
			name->scope[n++] = (char)ubuf[offset++];

		/*
		 * Watch for malicious loops.
		 */
		if (loop_count++ == 10)
			return 0;
	}
	name->scope[n++] = 0;

	return ret;
}

int put_nmb_name(char *buf, int offset, struct nmb_name *name)
{
	int ret,m;
	char buf1[128];
	char *p;

	if (strcmp(name->name,"*") == 0) {
		/* special case for wildcard name */
		memset(buf1,'\0',20);
		buf1[0] = '*';
		buf1[15] = name->name_type;
	} else {
		snprintf(buf1, sizeof(buf1) - 1,"%-15.15s%c",name->name,name->name_type);
	}

	buf[offset] = 0x20;
	ret = 34;

	for (m=0; m<16; m++) {
		buf[offset + 1 + 2*m] = 'A' + ((buf1[m] >> 4) & 0xF);
		buf[offset + 2 + 2*m] = 'A' + ( buf1[m]       & 0xF);
	}
	offset += 33;

	buf[offset] = 0;

	if (name->scope[0]) {
		/* XXXX this scope handling needs testing */
		ret += strlen(name->scope) + 1;
		strcpy(&buf[offset + 1], name->scope);

		p = &buf[offset + 1];
		while ((p = strchr(p, '.'))) {
			buf[offset] = PTR_DIFF(p, &buf[offset]);
			offset += buf[offset];
			p = &buf[offset+1];
		}
		buf[offset] = strlen(&buf[offset+1]);
	}

	return ret;
}

int handle_name_ptrs(unsigned char *ubuf, int *offset, int length,
                             int *got_pointer, int *ret)
{
	int loop_count=0;

	while ((ubuf[*offset] & 0xC0) == 0xC0) {
		if (!*got_pointer) (*ret) += 2;
		(*got_pointer) = 1;
		(*offset) = ((ubuf[*offset] & ~0xC0) << 8) | ubuf[(*offset)+1];
		if (loop_count++ == 10 || (*offset) < 0 || (*offset) > (length-2)) {
			return 0;
		}
	}
	return 1;
}

int parse_alloc_res_rec(char *inbuf, int *offset, int length,
			struct res_rec **recs, int count)
{
	int i;
	*recs = (struct res_rec *)malloc(sizeof(**recs) * count);
	if (!*recs) return 0;

	memset((char *)*recs, 0, sizeof(**recs) * count);

	for (i=0; i<count; i++) {
		int l = parse_nmb_name(inbuf, *offset, length, &(*recs)[i].rr_name);
		(*offset) += l;
		if (!l || (*offset)+10 > length) {
			free(*recs);
			*recs = NULL;
			return 0;
		}
		(*recs)[i].rr_type = RSVAL(inbuf, (*offset));
		(*recs)[i].rr_class = RSVAL(inbuf, (*offset)+2);
		(*recs)[i].ttl = RIVAL(inbuf, (*offset)+4);
		(*recs)[i].rdlength = RSVAL(inbuf, (*offset)+8);
		(*offset) += 10;
		if ((*recs)[i].rdlength > sizeof((*recs)[i].rdata) ||
		    (*offset)+(*recs)[i].rdlength > length) {
			free(*recs);
			*recs = NULL;
			return 0;
		}
		memcpy((*recs)[i].rdata, inbuf+(*offset), (*recs)[i].rdlength);
		(*offset) += (*recs)[i].rdlength;
	}
	return 0;
}

void free_nmb_packet(struct nmb_packet *nmb)
{
	if (nmb->answers) {
		free(nmb->answers);
		nmb->answers = NULL;
	}
	if (nmb->nsrecs) {
		free(nmb->nsrecs);
		nmb->nsrecs = NULL;
	}
	if (nmb->additional) {
		free(nmb->additional);
		nmb->additional = NULL;
	}
	free(nmb);
}

