/*
 * nbtscanner -- a tool for scanning large networks for SMB servers.
 *
 * socket.c: Most functions that are related to TCP/IP (socket) use.
 * Copyright (C) 2000 Steinar H. Gunderson
 *
 * Large amounts of code adapted from Samba (http://www.samba.org/)
 * Copyright (C) Andrew Tridgell 1994-1998, and others.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file is probably the most platform-dependent one, and would most
 * likely be needed to undergo some changes if ported to another OS than
 * the ones using BSD-style sockets (Winsock is a variant of these,
 * by the way -- you may only want to change some header files and add some
 * extra initialization calls).
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "socket.h"
#include "nmb.h"
#include "packet.h"

int open_sockets(void)
{
	unsigned int one = 1;
	struct sockaddr_in sin;
	int sock;

	memset((char *)&sin, 0, sizeof(sin));
	sin.sin_port = htons(137);
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = INADDR_ANY;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (sock == -1) {
		perror("socket()");
		exit(0);
	}

	setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(one));

	/* bind the socket */
	if (bind(sock, (struct sockaddr *) &sin,sizeof(sin)) < 0) {
		perror("bind()");
		exit(1);
	}

	return sock;
}

int send_packet(struct nmb_packet *nmb, struct in_addr ip, int port, int fd)
{
	char buf[1024];
	int len = 0;

	memset(buf, 0, sizeof(buf));

	len = build_nmb(buf,nmb);
	if (!len) return(0);

	return(send_udp(fd,buf,len,ip,port));
}

struct nmb_packet *receive_packet(int fd, int t, struct in_addr *in)
{
	fd_set fds;
	struct timeval timeout;

	FD_ZERO(&fds);
	FD_SET(fd,&fds);
	timeout.tv_sec = t / 1000;
	timeout.tv_usec = 1000 * (t % 1000);

	select(fd+1,&fds,NULL,NULL,&timeout);

	if (FD_ISSET(fd,&fds))
		return read_packet(fd, in);

	return(NULL);
}

struct nmb_packet *read_packet(int fd, struct in_addr *in)
{
	/* extern struct in_addr lastip;
	   extern int lastport; */
	char buf[MAX_DGRAM_SIZE];
	int length;
	int ok = 0;
	struct nmb_packet *nmb;

	length = read_udp_socket(fd, buf, sizeof(buf), in);
	if (length < MIN_DGRAM_SIZE) return(NULL);

	nmb = (struct nmb_packet *)malloc(sizeof(struct nmb_packet)); /* ?? */
	ok = parse_nmb(buf, length, nmb);

	return(nmb);
}

int send_udp(int fd,char *buf,int len,struct in_addr ip,int port)
{
	int ret;
	struct sockaddr_in sock_out;

	/* set the address and port */
	memset((char *)&sock_out, 0, sizeof(sock_out));
	memcpy((char *)&sock_out.sin_addr, (char *)&ip, 4);
	sock_out.sin_port = htons(port);
	sock_out.sin_family = AF_INET;

	ret = (sendto(fd,buf,len,0,(struct sockaddr *)&sock_out,
		sizeof(sock_out)) >= 0);

	return ret;
}

ssize_t read_udp_socket(int fd, char *buf, size_t len, struct in_addr *in)
{
	ssize_t ret;
	struct sockaddr_in sin;
	int sin_len;

	sin_len = sizeof(sin);
	memset((char *)&sin, 0, sin_len);
	ret = (ssize_t)recvfrom(fd, buf, len, 0, (struct sockaddr *)&sin, &sin_len);
	if (ret <= 0) {
		perror("recvfrom()");
		return(0);
	}

	*in = sin.sin_addr;

	return ret;
}


